-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.type = "auxilliary"
M.relevantDevice = rearLeftMotor, rearRightMotor, frontLeftMotor, frontRightMotor;

M.throttle_FL = 0
M.throttle_FR = 0
M.throttle_RL = 0
M.throttle_RR = 0

M.Debug = {}
M.Debug.FL = 0
M.Debug.FR = 0
M.Debug.RL = 0
M.Debug.RR = 0


--shorter functions for increased performance
local abs = math.abs
local min = math.min
local max = math.max
local sqrt = math.sqrt
local acos = math.acos
local clamp = clamp
local pi = math.pi
local pow = math.pow



-- debugging functions
local function printMembers(o)
	
	for key,value in pairs(o) do
		print(key .. " " .. value);
	end
	
end 

local function length(array)
	local c = 0
	for _ in pairs(array) do c = c + 1 end
	return c
end


local function average(values)
	
	local sum = 0
	
	for k,v in pairs(values) do
		sum = sum + v
	end
	
	return sum / length(values)
	
end


-- caching these because it's easier to read
-- also a stored reference has got to be faster than a string search
local motor_FL = powertrain.getDevice("frontLeftMotor")
local motor_FR = powertrain.getDevice("frontRightMotor")
local motor_RL = powertrain.getDevice("rearLeftMotor")
local motor_RR = powertrain.getDevice("rearRightMotor")

local wheel_RL = nil
local wheel_RR = nil
local wheel_FL = nil
local wheel_FR = nil


local skewStiffnessFront = 142000
local skewStiffnessRear = 177000
local distanceCOGFrontAxle = 0
local distanceCOGRearAxle = 0

local characteristicSpeed = 1
local invSquaredCharacteristicSpeed = 0 --pre calculated, used for desired yaw calculation

-- pid steering will be disabled if all these cutoffs are satisfied
local cutoffMinSpeed = 10/3.6 -- m/s, not km/h
local cutoffMinGForce = 0.2
local cutoffYawRate = 0.1

local configData = nil
local mass = 0

local wheelBase = 0
local invWheelBase = 0

local yawSmooth = nil
local steerSmooth = nil
local speed = 0
local desiredYawRateSteering = 0
local xGForce = 0


local crossAxleMaxFactor = 1 -- how much influence the cross-axle vectoring has 
local interAxleSteerFactor = 1 -- how much influence the inter-axle vectoring has
local interAxleGForcePower = 1/2 -- how the influence should be shaped by g-force
local interAxleCriticalGPoint = 10 -- where the 1,1 point lies on the power curve
local interAxleCriticalGPointInverse = 0

local throttleShapingPower = 1

local maxThrottleDivergence = 0.5 -- how much the throttle output values can depart from the throttle input

local skidSteerStationaryYawRate = 10 -- how fast the vehicle can rotate in skid-steer mode at a standstill
local skidSteerBoost = 1 -- how much to add to the fast yaw rate in skid-steer mode


local throttleDirection = 0


-- yer basic PID settings
local yaw_PID_gain = 1 -- strength
local yaw_P = 6 -- aggression
local yaw_I = 1 -- error correction
local yaw_D = 0.01 -- damping

-- yer basic PID object
local yaw_PID = {G = yaw_PID_gain, P = yaw_P, I = yaw_I, D = yaw_D, SP = 0, MPV = 0, MPV_prev = 0, I_sum = 0}


-- yer basic PID loop
local function update_PID(pid, sp, mpv, dt)
	
	pid.MPV_prev = pid.MPV
	pid.MPV = mpv
	pid.SP = sp
	
	local D_rate = (pid.MPV - pid.MPV_prev)/dt
	local P_error = pid.SP - pid.MPV
	pid.I_sum = pid.I_sum + P_error * dt
	
	return pid.G * (pid.P * P_error + pid.I * pid.I_sum - pid.D * D_rate)
	
end

-- this is to reset the PID integral whenever we change between forward, reverse and neutral
-- this is set manually and not from electrics.values.gear so that it can be overridden manually
local function setThrottleDirection(dir)
	
	if throttleDirection ~= dir then
		yaw_PID.I_sum = 0
		throttleDirection = dir
	end
	
end


local function cutoffCheck(speed, gforce, yaw, throttle, brake, parking)
	
		if speed < cutoffMinSpeed and gforce < cutoffMinGForce and yaw < cutoffYawRate and throttle == 0 and parking == 0 then
			return true
			
		elseif speed < cutoffMinSpeed and brake == 1 and parking == 0 then
			return true
		
		end
		
		return false
end


local function applyMotorThrottle(t, motor)
		
		-- reverse motor if throttle value is negative
		motor.motorDirection = fsign(t)
		-- set throttle to positive
		t = abs(t)
		
		-- return throttle and enjoy
		return t
	
end


local function normalisePower(values, p, throttle)
	
	local maxT = -1
	local minT = 1
	
	local tHeadroom = min(1 - throttle, maxThrottleDivergence)
	local tFootroom = min(1 + throttle, maxThrottleDivergence)
	
	for _,v in pairs(values) do
		maxT = max(v, maxT)
		minT = min(v, minT)
	end
	
	
	for k,_ in pairs(values) do
	
	
		if values[k] > throttle then
			
			if maxT > throttle + tHeadroom then
				values[k] = throttle + (values[k] - throttle) / (maxT - throttle) * tHeadroom
				
			end
			
			
		elseif values[k] < throttle then
			
			if minT < throttle - tFootroom then
				values[k] = throttle - (throttle - values[k]) / (throttle - minT) * tFootroom
				
			end
			
		end
		
		if values[k] > throttle then
			values[k] = throttle + pow(abs(values[k] - throttle), p) * tHeadroom
			
		elseif values[k] < throttle then
			values[k] = throttle - pow(abs(throttle - values[k]), p) * tFootroom
			
		end
			
			
		
	end
	
	
	-- local adjustSum = 0
	
	-- for _,v in pairs(values) do
		-- adjustSum = adjustSum + v - throttle
	-- end
	
	-- print(maxT)
	-- print(values.RR)
	
end



local function interAxleSteering(values, throttle, steer)
	
	
	
	for k,v in pairs(values) do
		
		if ((k == "FL" or k == "FR") and steer > 0) or ((k == "RL" or k == "RR") and steer < 0) then
			values[k] = v * (1 + abs(steer) * (1 - abs(v)))
		else
			values[k] = v / (1 + abs(steer))
			
		end
	end
	
end



local function updateWheelsIntermediate(dt)
	
	if electrics.values.gear == "N" or electrics.values.gear == "P" then
		setThrottleDirection(0)
		
		electrics.values.throttle_FL = 0
		electrics.values.throttle_FR = 0
		electrics.values.throttle_RL = 0
		electrics.values.throttle_RR = 0
	end
	
	
	local throttleInput = electrics.values.throttle_input
	local brakeInput = electrics.values.brake_input
	local parkingInput = electrics.values.parkingbrake_input
	
	-- getting these smoothed values here so there's no discontinuity in data collection for the smoothers
	local yawRate = yawSmooth:get(obj:getYawAngularVelocity())
	local steeringInput = steerSmooth:get(input.steering)
	
	if electrics.values.gear == "D" or electrics.values.gear == "R" then
		
		if electrics.values.gear == "R" then
			setThrottleDirection(-1)
			
		elseif electrics.values.gear == "D" then
			throttleInput = throttleInput - electrics.values.clutch_input
			
			if throttleInput > 0 then
				setThrottleDirection(1)
				
			elseif throttleInput < 0 then
				setThrottleDirection(-1)
				throttleInput = abs(throttleInput)
				
			end
			
		end
		
		
		-- desired yawrate will be based on wheelbase and speed
		local speed = electrics.values.wheelspeed
		local wheelFront = steeringInput < 0 and wheel_FL or wheel_FR
		local wheelRear = steeringInput < 0 and wheel_RL or wheel_RR
		
		
		-- stock ESC steering magic
		local wheelAngleFront = acos(obj:nodeVecPlanarCosRightForward(wheelFront.node1, wheelFront.node2))
		local wheelAngleRear = acos(obj:nodeVecPlanarCosRightForward(wheelRear.node1, wheelRear.node2))
		if wheelAngleFront > 1.5708 then
			wheelAngleFront = (pi - wheelAngleFront)
		end
		wheelAngleFront = wheelAngleFront * fsign(-steeringInput)
		if wheelAngleFront ~= wheelAngleFront then
			wheelAngleFront = 0
		end

		if wheelAngleRear > 1.5708 then
			wheelAngleRear = (pi - wheelAngleRear)
		end
		wheelAngleRear = wheelAngleRear * fsign(-steeringInput)
		if wheelAngleRear ~= wheelAngleRear then
			wheelAngleRear = 0
		end

		M.wheelAngleFront = wheelAngleFront
		M.wheelAngleRear = wheelAngleRear
		
		
		-- calculate expected yaw rate based on steering angle
		-- need to invert this when reversing because the PID doesn't like reversing its output
		desiredYawRateSteering = -((wheelAngleFront * invWheelBase) * (speed/(1 + (speed * speed * invSquaredCharacteristicSpeed)))) * throttleDirection
		
		
		-- parking brake maps to skid-steer mode when in drive
		if parkingInput > 0 then
			-- we don't want to use the actual parking brake and this seems to intercept it before it reaches the actual wheels and I'm not asking questions at this point
			input.parkingbrake = 0
			
			-- pick the stationary yaw rate, or the boosted fast yaw rate, whichever is higher
			desiredYawRateSteering = max(abs(skidSteerStationaryYawRate * parkingInput * steeringInput), abs(desiredYawRateSteering * (1 + skidSteerBoost * parkingInput))) * fsign(desiredYawRateSteering)
			
		end
		
		
		-- adjust the PIDSteer based on speed
		-- to get rid of juddering at lower speeds
		-- and improve response at high speeds
		local PIDSpeedFactor = (speed / characteristicSpeed)
		
		-- need to invert it again at this point because we lied to the PID, so it gives us an inverted reponse back
		local PIDOutput = update_PID(yaw_PID, desiredYawRateSteering, yawRate, dt)
		local PIDSteer = PIDOutput * max(PIDSpeedFactor, fsign(parkingInput)) * throttleDirection
		
		
		-- reduce cross-axle vectoring at higher speeds
		local crossAxleDelta = PIDSteer * crossAxleMaxFactor
		
		-- inter axle scaling is based entirely on lateral g-force and the desired steering results
		-- no need to look at steering angle or understeer/oversteer which are misleading names to say the least
		-- this is a scale because just putting a negative or positive throttle won't necessarily
		-- improve traction, we need to reduce or boost the power overall regardless of direction
		local xGForce = sensors.gx2
		
		
		if cutoffCheck(speed, xGForce, yawRate, throttleInput, brakeInput, parkingInput) then
			
			yaw_PID.I_sum = 0
			return
			
		end
		
		
		
		local interAxleSteerGFactor = pow(interAxleCriticalGPointInverse * abs(xGForce), interAxleGForcePower)
		local interAxleSteer = PIDSteer * interAxleSteerFactor * interAxleSteerGFactor
		
		--crossAxleDelta = crossAxleDelta-- * interAxleSteerGFactor
		
		-- negating this number switches from oversteer to understeer
		if fsign(xGForce) * throttleDirection < 0 then
			interAxleSteer = -interAxleSteer			
		end
		
		
		-- apply steering factors and reversing
		
		local throttleResults = {}
		throttleResults.FL = (throttleInput + crossAxleDelta) * throttleDirection
		throttleResults.FR = (throttleInput - crossAxleDelta) * throttleDirection
		throttleResults.RL = (throttleInput + crossAxleDelta) * throttleDirection
		throttleResults.RR = (throttleInput - crossAxleDelta) * throttleDirection
		
		
		normalisePower(throttleResults, throttleShapingPower, throttleInput * throttleDirection)
		
		
		interAxleSteering(throttleResults, throttleInput, interAxleSteer)
		
		--print(average(throttleResults))
		
		M.Debug.FL = throttleResults.FL
		M.Debug.FR = throttleResults.FR
		M.Debug.RL = throttleResults.RL
		M.Debug.RR = throttleResults.RR
		
		-- apply all throttle values
		electrics.values.throttle_FL = applyMotorThrottle(throttleResults.FL, motor_FL)
		electrics.values.throttle_FR = applyMotorThrottle(throttleResults.FR, motor_FR)
		electrics.values.throttle_RL = applyMotorThrottle(throttleResults.RL, motor_RL)		
		electrics.values.throttle_RR = applyMotorThrottle(throttleResults.RR, motor_RR)
		
		
	end
	
	
	
	
	
end

local function calculateCharacteristicSpeed()
	local eg = (mass * (skewStiffnessRear * distanceCOGRearAxle - skewStiffnessFront * distanceCOGFrontAxle)) / (skewStiffnessFront * skewStiffnessRear * wheelBase)
	local characteristicSpeed = sqrt(wheelBase / abs(eg + 1e-30)) --guard against infinity
	return characteristicSpeed
end


local function init(jbeamData)
  configData = jbeamData
end

local function initSecondStage()
	
	-- grab all the wheels and cache them
	for _,wd in pairs(wheels.wheels) do
		if (wd.name ==  "FL") then
			wheel_FL = wd
		elseif (wd.name == "FR") then
			wheel_FR = wd
		elseif (wd.name == "RL") then
			wheel_RL = wd
		elseif (wd.name == "RR") then
			wheel_RR = wd
		end
	end
	
	wheelBase = (vec3(v.data.nodes[wheel_FL.node1].pos) - vec3(v.data.nodes[wheel_RL.node1].pos)):length()
	
	invWheelBase = 1/wheelBase
	
	interAxleCriticalGPointInverse = 1 / interAxleCriticalGPoint
	
	yawSmooth = newExponentialSmoothing(50)
	
	steerSmooth = newExponentialSmoothing(10)
	
	local tmp = vec3(0, 0, 0)
	local totalMass = 0
	for _, v in pairs(v.data.nodes) do
		tmp = tmp + vec3(v.pos) * v.nodeWeight
		totalMass = totalMass + v.nodeWeight
	end

	local realCOG = tmp / totalMass
	
	mass = totalMass
	
	--Find the positions of the front and rear axle
	local frontAxlePos = 0
	local rearAxlePos = 0
	local twLeft = 0
	local twRight = 0
	for _, n in pairs(v.data.nodes) do
		if n.cid == wheel_FR.node1 then
			frontAxlePos = n.pos.y
			twRight = n.pos.x
		elseif n.cid == wheel_RR.node1 then
			rearAxlePos = n.pos.y
		elseif n.cid == wheel_FR.node1 then
			twLeft = n.pos.x
		end
	end
	
	
	
	distanceCOGFrontAxle = abs(realCOG.y - frontAxlePos)
	distanceCOGRearAxle = abs(realCOG.y - rearAxlePos)
	
	characteristicSpeed = calculateCharacteristicSpeed()
	
	invSquaredCharacteristicSpeed = 1 / (characteristicSpeed * characteristicSpeed)
	
	M.updateWheelsIntermediate = updateWheelsIntermediate
	M.updateGFX = updateGFX
end

local function shiftToGearIndex(index)
	print(index)
end

-- public interface
M.init = init
M.initSecondStage = initSecondStage
M.updateWheelsIntermediate = nil
M.updateGFX = nil
M.shiftToGearIndex = shiftToGearIndex

M.wheel_FL = wheel_FL

return M